// Copyright 2015-2019 Espressif Systems (Shanghai) PTE LTD
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at

//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
#include <string.h>
#include <sys/param.h>
#include "esp_types.h"
#include "esp_attr.h"
#include "esp_intr_alloc.h"
#include "esp_log.h"
#include "esp_err.h"
#include "uart_hal.h"
#include "hal/gpio_hal.h"
#include "soc/uart_periph.h"
#include "soc/rtc_cntl_reg.h"
#include "gpio.h"
#include "uart_select.h"
#include "periph_ctrl.h"
#include "sdkconfig.h"
#include "esp_rom_gpio.h"
#include "abup_custom_uart.h"
#include "abup_custom_timer.h"
#include "hal/cpu_hal.h"
#include "hal/clk_gate_ll.h"
#include "hal/interrupt_controller_hal.h"


#if CONFIG_IDF_TARGET_ESP32
#include "esp32/clk.h"
#elif CONFIG_IDF_TARGET_ESP32S2
#include "esp32s2/clk.h"
#elif CONFIG_IDF_TARGET_ESP32S3
#include "esp32s3/clk.h"
#elif CONFIG_IDF_TARGET_ESP32C3
#include "esp32c3/clk.h"
#elif CONFIG_IDF_TARGET_ESP32H2
#include "esp32h2/clk.h"
#endif

#ifdef CONFIG_UART_ISR_IN_IRAM
#define UART_ISR_ATTR IRAM_ATTR
#else
#define UART_ISR_ATTR
#endif

#define XOFF (0x13)
#define XON (0x11)

static const char* ABUP_UART_TAG = "abup_custom_uart";
#define UART_CHECK(a, str, ret_val) \
    if (!(a)) { \
        ESP_LOGE(ABUP_UART_TAG,"%s(%d): %s", __FUNCTION__, __LINE__, str); \
        return (ret_val); \
    }

#define ABUP_UART_EMPTY_THRESH_DEFAULT       (15)
#define ABUP_UART_FULL_THRESH_DEFAULT        (120)
#define ABUP_UART_TOUT_THRESH_DEFAULT        (15)
#define ABUP_UART_CLKDIV_FRAG_BIT_WIDTH      (3)
#define ABUP_UART_TX_IDLE_NUM_DEFAULT        (0)
#define ABUP_UART_PATTERN_DET_QLEN_DEFAULT   (10)
#define ABUP_UART_MIN_WAKEUP_THRESH          (UART_LL_MIN_WAKEUP_THRESH)

#define ABUP_UART_INTR_CONFIG_FLAG ((UART_INTR_RXFIFO_FULL) \
                            | (UART_INTR_RXFIFO_TOUT) \
                            | (UART_INTR_RXFIFO_OVF) \
                            | (UART_INTR_BRK_DET) \
                            | (UART_INTR_PARITY_ERR))


// Check actual UART mode set
#define UART_IS_MODE_SET(uart_number, mode) ((abup_uart_obj[uart_number].uart_mode == mode))

#define UART_CONTEX_INIT_DEF(uart_num) {\
    .hal.dev = UART_LL_GET_HW(uart_num),\
    .hw_enabled = false,\
}

#if SOC_UART_SUPPORT_RTC_CLK
#define RTC_ENABLED(uart_num)    (BIT(uart_num))
#endif

typedef struct {
    uart_event_type_t type;        /*!< UART TX data type */
    struct {
        int brk_len;
        size_t size;
        uint8_t data[0];
    } tx_data;
} uart_tx_data_t;

typedef struct {
    int wr;
    int rd;
    int len;
    int* data;
} uart_pat_rb_t;

typedef struct {
    uart_port_t uart_num;               /*!< UART port number*/
    int event_queue_size;               /*!< UART event queue size*/
    intr_handle_t intr_handle;          /*!< UART interrupt handle*/
    uart_mode_t uart_mode;              /*!< UART controller actual mode set by uart_set_mode() */
    bool coll_det_flg;                  /*!< UART collision detection flag */
    bool rx_always_timeout_flg;         /*!< UART always detect rx timeout flag */
    int rx_buffered_len;                /*!< UART cached data length */
    int rx_buf_size;                    /*!< RX ring buffer size */
    bool rx_buffer_full_flg;            /*!< RX ring buffer full flag. */
    uint32_t rx_cur_remain;             /*!< Data number that waiting to be read out in ring buffer item*/
    uint8_t *rx_ptr;                    /*!< pointer to the current data in ring buffer*/
    uint8_t *rx_head_ptr;               /*!< pointer to the head of RX item*/
    uint8_t rx_data_buf[SOC_UART_FIFO_LEN]; /*!< Data buffer to stash FIFO data*/
    uint8_t rx_stash_len;               /*!< stashed data length.(When using flow control, after reading out FIFO data, if we fail to push to buffer, we can just stash them.) */
    uint32_t rx_int_usr_mask;           /*!< RX interrupt status. Valid at any time, regardless of RX buffer status. */
    uart_pat_rb_t rx_pattern_pos;
    int tx_buf_size;                    /*!< TX ring buffer size */
    bool tx_waiting_fifo;               /*!< this flag indicates that some task is waiting for FIFO empty interrupt, used to send all data without any data buffer*/
    uint8_t *tx_ptr;                    /*!< TX data pointer to push to FIFO in TX buffer mode*/
    uart_tx_data_t *tx_head;            /*!< TX data pointer to head of the current buffer in TX ring buffer*/
    uint32_t tx_len_tot;                /*!< Total length of current item in ring buffer*/
    uint32_t tx_len_cur;
    uint8_t tx_brk_flg;                 /*!< Flag to indicate to send a break signal in the end of the item sending procedure */
    uint8_t tx_brk_len;                 /*!< TX break signal cycle length/number */
    uint8_t tx_waiting_brk;             /*!< Flag to indicate that TX FIFO is ready to send break signal after FIFO is empty, do not push data into TX FIFO right now.*/
    void * uart_select_notif_callback; /*!< Notification about select() events */
    void *event_queue;          /*!< UART event queue handler*/
#if CONFIG_UART_ISR_IN_IRAM
    void *event_queue_storage;
    void *event_queue_struct;
    void *rx_ring_buf_storage;
    void *rx_ring_buf_struct;
    void *tx_ring_buf_storage;
    void *tx_ring_buf_struct;
    void *rx_mux_struct;
    void *tx_mux_struct;
    void *tx_fifo_sem_struct;
    void *tx_done_sem_struct;
    void *tx_brk_sem_struct;
#endif
} abup_uart_obj_t;

typedef struct {
    uart_hal_context_t hal;        /*!< UART hal context*/
    bool hw_enabled;
} abup_uart_context_t;

static abup_uart_obj_t abup_uart_obj[UART_NUM_MAX] = {0};

static abup_uart_context_t abup_uart_context[UART_NUM_MAX] = {
    UART_CONTEX_INIT_DEF(UART_NUM_0),
    UART_CONTEX_INIT_DEF(UART_NUM_1),
#if UART_NUM_MAX > 2
    UART_CONTEX_INIT_DEF(UART_NUM_2),
#endif
};

#if SOC_UART_SUPPORT_RTC_CLK

static uint8_t rtc_enabled = 0;

static void rtc_clk_enable(uart_port_t uart_num)
{

    if (!(rtc_enabled & RTC_ENABLED(uart_num))) {
        rtc_enabled |= RTC_ENABLED(uart_num);
    }
    SET_PERI_REG_MASK(RTC_CNTL_CLK_CONF_REG, RTC_CNTL_DIG_CLK8M_EN_M);

}

static void rtc_clk_disable(uart_port_t uart_num)
{
    assert(rtc_enabled & RTC_ENABLED(uart_num));


    rtc_enabled &= ~RTC_ENABLED(uart_num);
    if (rtc_enabled == 0) {
        CLEAR_PERI_REG_MASK(RTC_CNTL_CLK_CONF_REG, RTC_CNTL_DIG_CLK8M_EN_M);
    }

}
#endif


static int UART_ISR_ATTR abup_uart_find_pattern_from_last(uint8_t* buf, int length, uint8_t pat_chr, uint8_t pat_num)
{
    int cnt = 0;
    int len = length;
    while (len >= 0) {
        if (buf[len] == pat_chr) {
            cnt++;
        } else {
            cnt = 0;
        }
        if (cnt >= pat_num) {
            break;
        }
        len --;
    }
    return len;
}

static void UART_ISR_ATTR abup_uart_group_isr_callback(void *param)
{
    abup_uart_obj_t *p_uart = (abup_uart_obj_t*) param;
    uint8_t uart_num = p_uart->uart_num;
    int rx_fifo_len = 0;
    uint32_t uart_intr_status = 0;
    uart_event_t uart_event;
    static uint8_t pat_flg = 0;

    while(1) {

        // The `continue statement` may cause the interrupt to loop infinitely
        // we exit the interrupt here
        uart_intr_status = uart_hal_get_intsts_mask(&(abup_uart_context[uart_num].hal));
        //Exit form while loop
        if(uart_intr_status == 0){
        //ESP_LOGI(ABUP_UART_TAG, "uart_intr_status == 0");
            break;
        }
        
        uart_event.type = UART_EVENT_MAX;
        if(uart_intr_status & UART_INTR_TXFIFO_EMPTY) {
            //ESP_LOGI(ABUP_UART_TAG, "abup_uart_group_isr_callback UART_INTR_TXFIFO_EMPTY");
            uart_hal_disable_intr_mask(&(abup_uart_context[uart_num].hal), UART_INTR_TXFIFO_EMPTY);
            uart_hal_clr_intsts_mask(&(abup_uart_context[uart_num].hal), UART_INTR_TXFIFO_EMPTY);
        }
        else if ((uart_intr_status & UART_INTR_RXFIFO_TOUT)
                || (uart_intr_status & UART_INTR_RXFIFO_FULL)
                || (uart_intr_status & UART_INTR_CMD_CHAR_DET)
                ) {
            if(pat_flg == 1) {
                uart_intr_status |= UART_INTR_CMD_CHAR_DET;
                pat_flg = 0;
            }
            if (p_uart->rx_buffer_full_flg == false) {
                //ESP_LOGI(ABUP_UART_TAG, "abup_uart_group_isr_callback rx_buffer_full_flg == false");
                rx_fifo_len = uart_hal_get_rxfifo_len(&(abup_uart_context[uart_num].hal));
                if ((abup_uart_obj[uart_num].rx_always_timeout_flg) && !(uart_intr_status & UART_INTR_RXFIFO_TOUT)) {
                    rx_fifo_len--; // leave one byte in the fifo in order to trigger uart_intr_rxfifo_tout
                }
                uart_hal_read_rxfifo(&(abup_uart_context[uart_num].hal), p_uart->rx_data_buf, &rx_fifo_len);
                uint8_t pat_chr = 0;
                uint8_t pat_num = 0;
                int pat_idx = -1;
                uart_hal_get_at_cmd_char(&(abup_uart_context[uart_num].hal), &pat_chr, &pat_num);

                //Get the buffer from the FIFO
                if (uart_intr_status & UART_INTR_CMD_CHAR_DET) {
                    uart_hal_clr_intsts_mask(&(abup_uart_context[uart_num].hal), UART_INTR_CMD_CHAR_DET);
                    uart_event.type = UART_PATTERN_DET;
                    uart_event.size = rx_fifo_len;
                    pat_idx = abup_uart_find_pattern_from_last(p_uart->rx_data_buf, rx_fifo_len - 1, pat_chr, pat_num);
                } else {
                    //After Copying the Data From FIFO ,Clear intr_status
                    uart_hal_clr_intsts_mask(&(abup_uart_context[uart_num].hal), UART_INTR_RXFIFO_TOUT | UART_INTR_RXFIFO_FULL);
                    uart_event.type = UART_DATA;
                    uart_event.size = rx_fifo_len;
                    uart_event.timeout_flag = (uart_intr_status & UART_INTR_RXFIFO_TOUT) ? true : false;
                }
                p_uart->rx_stash_len = rx_fifo_len;
            } else {
                //ESP_LOGI(ABUP_UART_TAG, "abup_uart_group_isr_callback rx_buffer_full_flg == true");
                uart_hal_disable_intr_mask(&(abup_uart_context[uart_num].hal), UART_INTR_RXFIFO_FULL | UART_INTR_RXFIFO_TOUT);

                uart_hal_clr_intsts_mask(&(abup_uart_context[uart_num].hal), UART_INTR_RXFIFO_FULL | UART_INTR_RXFIFO_TOUT);
                if(uart_intr_status & UART_INTR_CMD_CHAR_DET) {
                    uart_hal_clr_intsts_mask(&(abup_uart_context[uart_num].hal), UART_INTR_CMD_CHAR_DET);
                    uart_event.type = UART_PATTERN_DET;
                    uart_event.size = rx_fifo_len;
                    pat_flg = 1;
                }
            }
        } else if(uart_intr_status & UART_INTR_RXFIFO_OVF) {
            //ESP_LOGI(ABUP_UART_TAG, "abup_uart_group_isr_callback UART_INTR_RXFIFO_OVF");
            // When fifo overflows, we reset the fifo.
            uart_hal_rxfifo_rst(&(abup_uart_context[uart_num].hal));

            uart_hal_clr_intsts_mask(&(abup_uart_context[uart_num].hal), UART_INTR_RXFIFO_OVF);
            uart_event.type = UART_FIFO_OVF;
        } else if(uart_intr_status & UART_INTR_BRK_DET) {
        //ESP_LOGI(ABUP_UART_TAG, "abup_uart_group_isr_callback UART_INTR_BRK_DET");
            uart_hal_clr_intsts_mask(&(abup_uart_context[uart_num].hal), UART_INTR_BRK_DET);
            uart_event.type = UART_BREAK;
        } else if(uart_intr_status & UART_INTR_FRAM_ERR) {
        //ESP_LOGI(ABUP_UART_TAG, "abup_uart_group_isr_callback UART_INTR_FRAM_ERR");
            uart_hal_clr_intsts_mask(&(abup_uart_context[uart_num].hal), UART_INTR_FRAM_ERR);
            uart_event.type = UART_FRAME_ERR;
        } else if(uart_intr_status & UART_INTR_PARITY_ERR) {
        //ESP_LOGI(ABUP_UART_TAG, "abup_uart_group_isr_callback UART_INTR_PARITY_ERR");
            uart_hal_clr_intsts_mask(&(abup_uart_context[uart_num].hal), UART_INTR_PARITY_ERR);
            uart_event.type = UART_PARITY_ERR;
        } else if(uart_intr_status & UART_INTR_TX_BRK_DONE) {
        //ESP_LOGI(ABUP_UART_TAG, "abup_uart_group_isr_callback UART_INTR_TX_BRK_DONE");
            uart_hal_tx_break(&(abup_uart_context[uart_num].hal), 0);
            uart_hal_disable_intr_mask(&(abup_uart_context[uart_num].hal), UART_INTR_TX_BRK_DONE);
            if(p_uart->tx_brk_flg == 1) {
                uart_hal_ena_intr_mask(&(abup_uart_context[uart_num].hal), UART_INTR_TXFIFO_EMPTY);
            }

            uart_hal_clr_intsts_mask(&(abup_uart_context[uart_num].hal), UART_INTR_TX_BRK_DONE);
            if(p_uart->tx_brk_flg == 1) {
                p_uart->tx_brk_flg = 0;
                p_uart->tx_waiting_brk = 0;
            }
        } else if(uart_intr_status & UART_INTR_TX_BRK_IDLE) {
        //ESP_LOGI(ABUP_UART_TAG, "abup_uart_group_isr_callback UART_INTR_TX_BRK_IDLE");
            uart_hal_disable_intr_mask(&(abup_uart_context[uart_num].hal), UART_INTR_TX_BRK_IDLE);

            uart_hal_clr_intsts_mask(&(abup_uart_context[uart_num].hal), UART_INTR_TX_BRK_IDLE);
        } else if(uart_intr_status & UART_INTR_CMD_CHAR_DET) {
        //ESP_LOGI(ABUP_UART_TAG, "abup_uart_group_isr_callback UART_INTR_CMD_CHAR_DET");
            uart_hal_clr_intsts_mask(&(abup_uart_context[uart_num].hal), UART_INTR_CMD_CHAR_DET);
            uart_event.type = UART_PATTERN_DET;
        } else if(uart_intr_status & UART_INTR_TX_DONE) {
        //ESP_LOGI(ABUP_UART_TAG, "abup_uart_group_isr_callback UART_INTR_TX_DONE");
            if (UART_IS_MODE_SET(uart_num, UART_MODE_RS485_HALF_DUPLEX) && uart_hal_is_tx_idle(&(abup_uart_context[uart_num].hal)) != true) {
                // The TX_DONE interrupt is triggered but transmit is active
                // then postpone interrupt processing for next interrupt
                uart_event.type = UART_EVENT_MAX;
            } else {
                // Workaround for RS485: If the RS485 half duplex mode is active
                // and transmitter is in idle state then reset received buffer and reset RTS pin
                // skip this behavior for other UART modes
                uart_hal_disable_intr_mask(&(abup_uart_context[uart_num].hal), UART_INTR_TX_DONE);

                uart_hal_clr_intsts_mask(&(abup_uart_context[uart_num].hal), UART_INTR_TX_DONE);
            }
        } else {
            //ESP_LOGI(ABUP_UART_TAG, "abup_uart_group_isr_callback last else");
            uart_hal_clr_intsts_mask(&(abup_uart_context[uart_num].hal), uart_intr_status); /*simply clear all other intr status*/
            uart_event.type = UART_EVENT_MAX;
        }
    }
}


esp_err_t abup_boot_uart_init(void)
{
    esp_err_t ret;
    const uart_config_t abup_uart_cfg = {
        .baud_rate = 115200,   //1.5M
        .data_bits = UART_DATA_8_BITS,
        .parity = UART_PARITY_DISABLE,
        .stop_bits = UART_STOP_BITS_1,
        .flow_ctrl = UART_HW_FLOWCTRL_DISABLE,
        .rx_flow_ctrl_thresh = 127,
    };
    ESP_LOGI(ABUP_UART_TAG, "abup_boot_uart_init is running");
    ret =  abup_uart_param_config(ABUP_CONSOLE_UART_NUM, &abup_uart_cfg);
    if (ret != ESP_OK) {
        ESP_LOGE(ABUP_UART_TAG, "Uart %d initialize err %04x", ABUP_CONSOLE_UART_NUM, ret);
        return ret;
    }
    abup_uart_set_pin(ABUP_CONSOLE_UART_NUM, 4, 5, UART_PIN_NO_CHANGE, UART_PIN_NO_CHANGE);
    abup_uart_enable_pattern_det_baud_intr(ABUP_CONSOLE_UART_NUM, '\n', 1, 9, 0, 0);
    abup_uart_driver_install(ABUP_CONSOLE_UART_NUM, 512, 512, 10, 0);

    return ESP_OK;
}

static void abup_uart_module_enable(uart_port_t uart_num)
{
    if (abup_uart_context[uart_num].hw_enabled != true) {
        periph_module_enable(uart_periph_signal[uart_num].module);
        if (uart_num != CONFIG_ESP_CONSOLE_UART_NUM) {
            // Workaround for ESP32C3: enable core reset
            // before enabling uart module clock
            // to prevent uart output garbage value.
#if SOC_UART_REQUIRE_CORE_RESET
            uart_hal_set_reset_core(&(abup_uart_context[uart_num].hal), true);
            periph_module_reset(uart_periph_signal[uart_num].module);
            uart_hal_set_reset_core(&(abup_uart_context[uart_num].hal), false);
#else
            periph_module_reset(uart_periph_signal[uart_num].module);
#endif
        }
        abup_uart_context[uart_num].hw_enabled = true;
    }
}

static void uart_module_disable(uart_port_t uart_num)
{
    if (abup_uart_context[uart_num].hw_enabled != false) {
        if (uart_num != CONFIG_ESP_CONSOLE_UART_NUM ) {
            periph_module_disable(uart_periph_signal[uart_num].module);
        }
        abup_uart_context[uart_num].hw_enabled = false;
    }
}

esp_err_t uart_set_word_length(uart_port_t uart_num, uart_word_length_t data_bit)
{
    UART_CHECK((uart_num < UART_NUM_MAX), "uart_num error", ESP_FAIL);
    UART_CHECK((data_bit < UART_DATA_BITS_MAX), "data bit error", ESP_FAIL);

    uart_hal_set_data_bit_num(&(abup_uart_context[uart_num].hal), data_bit);

    return ESP_OK;
}

esp_err_t uart_get_word_length(uart_port_t uart_num, uart_word_length_t* data_bit)
{
    UART_CHECK((uart_num < UART_NUM_MAX), "uart_num error", ESP_FAIL);
    uart_hal_get_data_bit_num(&(abup_uart_context[uart_num].hal), data_bit);
    return ESP_OK;
}

esp_err_t uart_set_stop_bits(uart_port_t uart_num, uart_stop_bits_t stop_bit)
{
    UART_CHECK((uart_num < UART_NUM_MAX), "uart_num error", ESP_FAIL);
    UART_CHECK((stop_bit < UART_STOP_BITS_MAX), "stop bit error", ESP_FAIL);

    uart_hal_set_stop_bits(&(abup_uart_context[uart_num].hal), stop_bit);

    return ESP_OK;
}

esp_err_t uart_get_stop_bits(uart_port_t uart_num, uart_stop_bits_t* stop_bit)
{
    UART_CHECK((uart_num < UART_NUM_MAX), "uart_num error", ESP_FAIL);
    uart_hal_get_stop_bits(&(abup_uart_context[uart_num].hal), stop_bit);
    return ESP_OK;
}

esp_err_t uart_set_parity(uart_port_t uart_num, uart_parity_t parity_mode)
{
    UART_CHECK((uart_num < UART_NUM_MAX), "uart_num error", ESP_FAIL);

    uart_hal_set_parity(&(abup_uart_context[uart_num].hal), parity_mode);

    return ESP_OK;
}

esp_err_t uart_get_parity(uart_port_t uart_num, uart_parity_t* parity_mode)
{
    UART_CHECK((uart_num < UART_NUM_MAX), "uart_num error", ESP_FAIL);
    uart_hal_get_parity(&(abup_uart_context[uart_num].hal), parity_mode);
    return ESP_OK;
}

esp_err_t uart_set_baudrate(uart_port_t uart_num, uint32_t baud_rate)
{
    UART_CHECK((uart_num < UART_NUM_MAX), "uart_num error", ESP_FAIL);

    uart_hal_set_baudrate(&(abup_uart_context[uart_num].hal), baud_rate);

    return ESP_OK;
}

esp_err_t uart_get_baudrate(uart_port_t uart_num, uint32_t *baudrate)
{
    UART_CHECK((uart_num < UART_NUM_MAX), "uart_num error", ESP_FAIL);

    uart_hal_get_baudrate(&(abup_uart_context[uart_num].hal), baudrate);

    return ESP_OK;
}

esp_err_t uart_set_line_inverse(uart_port_t uart_num, uint32_t inverse_mask)
{
    UART_CHECK((uart_num < UART_NUM_MAX), "uart_num error", ESP_FAIL);

    uart_hal_inverse_signal(&(abup_uart_context[uart_num].hal), inverse_mask);

    return ESP_OK;
}

esp_err_t uart_set_sw_flow_ctrl(uart_port_t uart_num, bool enable,  uint8_t rx_thresh_xon,  uint8_t rx_thresh_xoff)
{
    UART_CHECK((uart_num < UART_NUM_MAX), "uart_num error", ESP_FAIL);
    UART_CHECK((rx_thresh_xon < SOC_UART_FIFO_LEN), "rx flow xon thresh error", ESP_FAIL);
    UART_CHECK((rx_thresh_xoff < SOC_UART_FIFO_LEN), "rx flow xoff thresh error", ESP_FAIL);
    uart_sw_flowctrl_t sw_flow_ctl = {
        .xon_char = XON,
        .xoff_char = XOFF,
        .xon_thrd = rx_thresh_xon,
        .xoff_thrd = rx_thresh_xoff,
    };

    uart_hal_set_sw_flow_ctrl(&(abup_uart_context[uart_num].hal), &sw_flow_ctl, enable);

    return ESP_OK;
}

esp_err_t uart_set_hw_flow_ctrl(uart_port_t uart_num, uart_hw_flowcontrol_t flow_ctrl, uint8_t rx_thresh)
{
    UART_CHECK((uart_num < UART_NUM_MAX), "uart_num error", ESP_FAIL);
    UART_CHECK((rx_thresh < SOC_UART_FIFO_LEN), "rx flow thresh error", ESP_FAIL);
    UART_CHECK((flow_ctrl < UART_HW_FLOWCTRL_MAX), "hw_flowctrl mode error", ESP_FAIL);

    uart_hal_set_hw_flow_ctrl(&(abup_uart_context[uart_num].hal), flow_ctrl, rx_thresh);

    return ESP_OK;
}

esp_err_t uart_get_hw_flow_ctrl(uart_port_t uart_num, uart_hw_flowcontrol_t* flow_ctrl)
{
    UART_CHECK((uart_num < UART_NUM_MAX), "uart_num error", ESP_FAIL)

    uart_hal_get_hw_flow_ctrl(&(abup_uart_context[uart_num].hal), flow_ctrl);

    return ESP_OK;
}

esp_err_t UART_ISR_ATTR uart_clear_intr_status(uart_port_t uart_num, uint32_t clr_mask)
{
    UART_CHECK((uart_num < UART_NUM_MAX), "uart_num error", ESP_FAIL);
    uart_hal_clr_intsts_mask(&(abup_uart_context[uart_num].hal), clr_mask);
    return ESP_OK;
}

esp_err_t uart_enable_intr_mask(uart_port_t uart_num, uint32_t enable_mask)
{
    UART_CHECK((uart_num < UART_NUM_MAX), "uart_num error", ESP_FAIL);

    abup_uart_obj[uart_num].rx_int_usr_mask |= enable_mask;
    uart_hal_clr_intsts_mask(&(abup_uart_context[uart_num].hal), enable_mask);
    uart_hal_ena_intr_mask(&(abup_uart_context[uart_num].hal), enable_mask);

    return ESP_OK;
}


esp_err_t uart_disable_intr_mask(uart_port_t uart_num, uint32_t disable_mask)
{
    UART_CHECK((uart_num < UART_NUM_MAX), "uart_num error", ESP_FAIL);

    abup_uart_obj[uart_num].rx_int_usr_mask &= ~disable_mask;
    uart_hal_disable_intr_mask(&(abup_uart_context[uart_num].hal), disable_mask);

    return ESP_OK;
}


static esp_err_t UART_ISR_ATTR uart_pattern_enqueue(uart_port_t uart_num, int pos)
{
    esp_err_t ret = ESP_OK;
    uart_pat_rb_t* p_pos = &abup_uart_obj[uart_num].rx_pattern_pos;
    int next = p_pos->wr + 1;
    if (next >= p_pos->len) {
        next = 0;
    }
    if (next == p_pos->rd) {
#ifndef CONFIG_UART_ISR_IN_IRAM     //Only log if ISR is not in IRAM
        ESP_EARLY_LOGW(ABUP_UART_TAG, "Fail to enqueue pattern position, pattern queue is full.");
#endif
        ret = ESP_FAIL;
    } else {
        p_pos->data[p_pos->wr] = pos;
        p_pos->wr = next;
        ret = ESP_OK;
    }
    return ret;
}

static esp_err_t uart_pattern_dequeue(uart_port_t uart_num)
{
    if(abup_uart_obj[uart_num].rx_pattern_pos.data == NULL) {
        return ESP_ERR_INVALID_STATE;
    } else {
        esp_err_t ret = ESP_OK;
        uart_pat_rb_t* p_pos = &abup_uart_obj[uart_num].rx_pattern_pos;
        if (p_pos->rd == p_pos->wr) {
            ret = ESP_FAIL;
        } else {
            p_pos->rd++;
        }
        if (p_pos->rd >= p_pos->len) {
            p_pos->rd = 0;
        }
        return ret;
    }
}

static esp_err_t uart_pattern_queue_update(uart_port_t uart_num, int diff_len)
{
    uart_pat_rb_t* p_pos = &abup_uart_obj[uart_num].rx_pattern_pos;
    int rd = p_pos->rd;
    while(rd != p_pos->wr) {
        p_pos->data[rd] -= diff_len;
        int rd_rec = rd;
        rd ++;
        if (rd >= p_pos->len) {
            rd = 0;
        }
        if (p_pos->data[rd_rec] < 0) {
            p_pos->rd = rd;
        }
    }
    return ESP_OK;
}

int uart_pattern_pop_pos(uart_port_t uart_num)
{

    uart_pat_rb_t* pat_pos = &abup_uart_obj[uart_num].rx_pattern_pos;
    int pos = -1;
    if (pat_pos != NULL && pat_pos->rd != pat_pos->wr) {
        pos = pat_pos->data[pat_pos->rd];
        uart_pattern_dequeue(uart_num);
    }

    return pos;
}

int uart_pattern_get_pos(uart_port_t uart_num)
{

    uart_pat_rb_t* pat_pos = &abup_uart_obj[uart_num].rx_pattern_pos;
    int pos = -1;
    if (pat_pos != NULL && pat_pos->rd != pat_pos->wr) {
        pos = pat_pos->data[pat_pos->rd];
    }

    return pos;
}

esp_err_t uart_pattern_queue_reset(uart_port_t uart_num, int queue_length)
{
    UART_CHECK((uart_num < UART_NUM_MAX), "uart_num error", ESP_FAIL);

    int* pdata = (int*) malloc(queue_length * sizeof(int));
    if(pdata == NULL) {
        return ESP_ERR_NO_MEM;
    }

    int* ptmp = abup_uart_obj[uart_num].rx_pattern_pos.data;
    abup_uart_obj[uart_num].rx_pattern_pos.data = pdata;
    abup_uart_obj[uart_num].rx_pattern_pos.len = queue_length;
    abup_uart_obj[uart_num].rx_pattern_pos.rd = 0;
    abup_uart_obj[uart_num].rx_pattern_pos.wr = 0;

    free(ptmp);
    return ESP_OK;
}

#if CONFIG_IDF_TARGET_ESP32
esp_err_t uart_enable_pattern_det_intr(uart_port_t uart_num, char pattern_chr, uint8_t chr_num, int chr_tout, int post_idle, int pre_idle)
{
    UART_CHECK((uart_num < UART_NUM_MAX), "uart_num error", ESP_FAIL);
    UART_CHECK(chr_tout >= 0 && chr_tout <= UART_RX_GAP_TOUT_V, "uart pattern set error\n", ESP_FAIL);
    UART_CHECK(post_idle >= 0 && post_idle <= UART_POST_IDLE_NUM_V, "uart pattern set error\n", ESP_FAIL);
    UART_CHECK(pre_idle >= 0 && pre_idle <= UART_PRE_IDLE_NUM_V, "uart pattern set error\n", ESP_FAIL);
    uart_at_cmd_t at_cmd = {0};
    at_cmd.cmd_char = pattern_chr;
    at_cmd.char_num = chr_num;
    at_cmd.gap_tout = chr_tout;
    at_cmd.pre_idle = pre_idle;
    at_cmd.post_idle = post_idle;
    uart_hal_clr_intsts_mask(&(abup_uart_context[uart_num].hal), UART_INTR_CMD_CHAR_DET);

    uart_hal_set_at_cmd_char(&(abup_uart_context[uart_num].hal), &at_cmd);
    uart_hal_ena_intr_mask(&(abup_uart_context[uart_num].hal), UART_INTR_CMD_CHAR_DET);

    return ESP_OK;
}
#endif

esp_err_t abup_uart_enable_pattern_det_baud_intr(uart_port_t uart_num, char pattern_chr, uint8_t chr_num, int chr_tout, int post_idle, int pre_idle)
{
    UART_CHECK((uart_num < UART_NUM_MAX), "uart_num error", ESP_FAIL);
    UART_CHECK(chr_tout >= 0 && chr_tout <= UART_RX_GAP_TOUT_V, "uart pattern set error\n", ESP_FAIL);
    UART_CHECK(post_idle >= 0 && post_idle <= UART_POST_IDLE_NUM_V, "uart pattern set error\n", ESP_FAIL);
    UART_CHECK(pre_idle >= 0 && pre_idle <= UART_PRE_IDLE_NUM_V, "uart pattern set error\n", ESP_FAIL);
    uart_at_cmd_t at_cmd = {0};
    at_cmd.cmd_char = pattern_chr;
    at_cmd.char_num = chr_num;

#if CONFIG_IDF_TARGET_ESP32
    int apb_clk_freq = 0;
    uint32_t uart_baud = 0;
    uint32_t uart_div = 0;
    uart_get_baudrate(uart_num, &uart_baud);
    apb_clk_freq = esp_clk_apb_freq();
    uart_div = apb_clk_freq / uart_baud;

    at_cmd.gap_tout = chr_tout * uart_div;
    at_cmd.pre_idle = pre_idle * uart_div;
    at_cmd.post_idle = post_idle * uart_div;
#else
    at_cmd.gap_tout = chr_tout;
    at_cmd.pre_idle = pre_idle;
    at_cmd.post_idle = post_idle;
#endif
    uart_hal_clr_intsts_mask(&(abup_uart_context[uart_num].hal), UART_INTR_CMD_CHAR_DET);

    uart_hal_set_at_cmd_char(&(abup_uart_context[uart_num].hal), &at_cmd);
    uart_hal_ena_intr_mask(&(abup_uart_context[uart_num].hal), UART_INTR_CMD_CHAR_DET);

    return ESP_OK;
}


esp_err_t uart_disable_pattern_det_intr(uart_port_t uart_num)
{
    return uart_disable_intr_mask(uart_num, UART_INTR_CMD_CHAR_DET);
}

esp_err_t abup_uart_enable_rx_intr(uart_port_t uart_num)
{
    return uart_enable_intr_mask(uart_num, UART_INTR_RXFIFO_FULL|UART_INTR_RXFIFO_TOUT);
}

esp_err_t uart_disable_rx_intr(uart_port_t uart_num)
{
    return uart_disable_intr_mask(uart_num, UART_INTR_RXFIFO_FULL|UART_INTR_RXFIFO_TOUT);
}

esp_err_t uart_disable_tx_intr(uart_port_t uart_num)
{
    return uart_disable_intr_mask(uart_num, UART_INTR_TXFIFO_EMPTY);
}

esp_err_t uart_enable_tx_intr(uart_port_t uart_num, int enable, int thresh)
{
    UART_CHECK((uart_num < UART_NUM_MAX), "uart_num error", ESP_FAIL);
    UART_CHECK((thresh < SOC_UART_FIFO_LEN), "empty intr threshold error", ESP_FAIL);
    if (enable == 0) {
    uart_hal_disable_intr_mask(&(abup_uart_context[uart_num].hal), UART_INTR_TXFIFO_EMPTY);
    } else {
        uart_hal_clr_intsts_mask(&(abup_uart_context[uart_num].hal), UART_INTR_TXFIFO_EMPTY);
    	uart_hal_set_txfifo_empty_thr(&(abup_uart_context[uart_num].hal), thresh);
    	uart_hal_ena_intr_mask(&(abup_uart_context[uart_num].hal), UART_INTR_TXFIFO_EMPTY);
    }
    return ESP_OK;
}


static esp_err_t abup_esp_intr_alloc_intrstatus(int source, intr_handler_t handler, void *arg)
{
    int intr_num = 13;//固定起始中断号13
    //ESP_LOGI(ABUP_TIMER_TAG, "esp_intr_alloc_intrstatus (cpu %d): checking args", cpu_hal_get_core_id());

    int cpu=cpu_hal_get_core_id();
    
    if (source>=0) {
        intr_matrix_set(cpu, source, intr_num);
    }
    
    if (handler) {
        interrupt_controller_hal_set_int_handler(intr_num, handler, arg);
    }

    //non_iram_int_mask[cpu]|=(1<<intr_num);

    //Enable int at CPU-level;
    ABUP_ESP_INTR_ENABLE(intr_num);

#ifdef SOC_CPU_HAS_FLEXIBLE_INTC
    //Extract the level from the interrupt passed flags
    int level = __builtin_ffs(0) + 1;
    interrupt_controller_hal_set_int_level(intr_num, level);
    interrupt_controller_hal_set_int_type(intr_num, INTTP_LEVEL);
#endif

    ESP_LOGI(ABUP_UART_TAG, "Connected src %d to int %d (cpu %d)", source, intr_num, cpu);
    return ESP_OK;
}

esp_err_t abup_uart_isr_register(uart_port_t uart_num, intr_handler_t handler, void*arg)
{
    int ret = ESP_FAIL;
    UART_CHECK((uart_num < UART_NUM_MAX), "uart_num error", ESP_FAIL);

    ret = abup_esp_intr_alloc_intrstatus(uart_periph_signal[uart_num].irq, handler, arg);

    return ret;
}


static bool uart_try_set_iomux_pin(uart_port_t uart_num, int io_num, uint32_t idx)
{
    /* Store a pointer to the default pin, to optimize access to its fields. */
    const uart_periph_sig_t *upin = &uart_periph_signal[uart_num].pins[idx];

    /* In theory, if default_gpio is -1, iomux_func should also be -1, but
     * let's be safe and test both. */
    if (upin->iomux_func == -1 || upin->default_gpio == -1 || upin->default_gpio != io_num) {
        return false;
    }

    /* Assign the correct funct to the GPIO. */
    assert (upin->iomux_func != -1);
    gpio_iomux_out(io_num, upin->iomux_func, false);

    /* If the pin is input, we also have to redirect the signal,
     * in order to bypasse the GPIO matrix. */
    if (upin->input) {
        gpio_iomux_in(io_num, upin->signal);
    }

    return true;
}

//internal signal can be output to multiple GPIO pads
//only one GPIO pad can connect with input signal
esp_err_t abup_uart_set_pin(uart_port_t uart_num, int tx_io_num, int rx_io_num, int rts_io_num, int cts_io_num)
{
    UART_CHECK((uart_num < UART_NUM_MAX), "uart_num error", ESP_FAIL);
    UART_CHECK((tx_io_num < 0 || (GPIO_IS_VALID_OUTPUT_GPIO(tx_io_num))), "tx_io_num error", ESP_FAIL);
    UART_CHECK((rx_io_num < 0 || (GPIO_IS_VALID_GPIO(rx_io_num))), "rx_io_num error", ESP_FAIL);
    UART_CHECK((rts_io_num < 0 || (GPIO_IS_VALID_OUTPUT_GPIO(rts_io_num))), "rts_io_num error", ESP_FAIL);
    UART_CHECK((cts_io_num < 0 || (GPIO_IS_VALID_GPIO(cts_io_num))), "cts_io_num error", ESP_FAIL);

    /* In the following statements, if the io_num is negative, no need to configure anything. */
    if (tx_io_num >= 0 && !uart_try_set_iomux_pin(uart_num, tx_io_num, SOC_UART_TX_PIN_IDX)) {
        gpio_hal_iomux_func_sel(GPIO_PIN_MUX_REG[tx_io_num], PIN_FUNC_GPIO);
        gpio_set_level(tx_io_num, 1);
        esp_rom_gpio_connect_out_signal(tx_io_num, UART_PERIPH_SIGNAL(uart_num, SOC_UART_TX_PIN_IDX), 0, 0);
    }

    if (rx_io_num >= 0 && !uart_try_set_iomux_pin(uart_num, rx_io_num, SOC_UART_RX_PIN_IDX)) {
        gpio_hal_iomux_func_sel(GPIO_PIN_MUX_REG[rx_io_num], PIN_FUNC_GPIO);
        gpio_set_pull_mode(rx_io_num, GPIO_PULLUP_ONLY);
        gpio_set_direction(rx_io_num, GPIO_MODE_INPUT);
        esp_rom_gpio_connect_in_signal(rx_io_num, UART_PERIPH_SIGNAL(uart_num, SOC_UART_RX_PIN_IDX), 0);
    }

    if (rts_io_num >= 0 && !uart_try_set_iomux_pin(uart_num, rts_io_num, SOC_UART_RTS_PIN_IDX)) {
        gpio_hal_iomux_func_sel(GPIO_PIN_MUX_REG[rts_io_num], PIN_FUNC_GPIO);
        gpio_set_direction(rts_io_num, GPIO_MODE_OUTPUT);
        esp_rom_gpio_connect_out_signal(rts_io_num, UART_PERIPH_SIGNAL(uart_num, SOC_UART_RTS_PIN_IDX), 0, 0);
    }

    if (cts_io_num >= 0  && !uart_try_set_iomux_pin(uart_num, cts_io_num, SOC_UART_CTS_PIN_IDX)) {
        gpio_hal_iomux_func_sel(GPIO_PIN_MUX_REG[cts_io_num], PIN_FUNC_GPIO);
        gpio_set_pull_mode(cts_io_num, GPIO_PULLUP_ONLY);
        gpio_set_direction(cts_io_num, GPIO_MODE_INPUT);
        esp_rom_gpio_connect_in_signal(cts_io_num, UART_PERIPH_SIGNAL(uart_num, SOC_UART_CTS_PIN_IDX), 0);
    }

    return ESP_OK;
}

esp_err_t uart_set_rts(uart_port_t uart_num, int level)
{
    UART_CHECK((uart_num < UART_NUM_MAX), "uart_num error", ESP_FAIL);
    UART_CHECK((!uart_hal_is_hw_rts_en(&(abup_uart_context[uart_num].hal))), "disable hw flowctrl before using sw control", ESP_FAIL);

    uart_hal_set_rts(&(abup_uart_context[uart_num].hal), level);

    return ESP_OK;
}

esp_err_t uart_set_dtr(uart_port_t uart_num, int level)
{
    UART_CHECK((uart_num < UART_NUM_MAX), "uart_num error", ESP_FAIL);

    uart_hal_set_dtr(&(abup_uart_context[uart_num].hal), level);

    return ESP_OK;
}

esp_err_t uart_set_tx_idle_num(uart_port_t uart_num, uint16_t idle_num)
{
    UART_CHECK((uart_num < UART_NUM_MAX), "uart_num error", ESP_FAIL);
    UART_CHECK((idle_num <= UART_TX_IDLE_NUM_V), "uart idle num error", ESP_FAIL);

    uart_hal_set_tx_idle_num(&(abup_uart_context[uart_num].hal), idle_num);

    return ESP_OK;
}

esp_err_t abup_uart_param_config(uart_port_t uart_num, const uart_config_t *uart_config)
{
    UART_CHECK((uart_num < UART_NUM_MAX), "uart_num error", ESP_FAIL);
    UART_CHECK((uart_config), "param null", ESP_FAIL);
    UART_CHECK((uart_config->rx_flow_ctrl_thresh < SOC_UART_FIFO_LEN), "rx flow thresh error", ESP_FAIL);
    UART_CHECK((uart_config->flow_ctrl < UART_HW_FLOWCTRL_MAX), "hw_flowctrl mode error", ESP_FAIL);
    UART_CHECK((uart_config->data_bits < UART_DATA_BITS_MAX), "data bit error", ESP_FAIL);
    abup_uart_module_enable(uart_num);
#if SOC_UART_SUPPORT_RTC_CLK
    if (uart_config->source_clk == UART_SCLK_RTC) {
        rtc_clk_enable(uart_num);
    }
#endif

    uart_hal_init(&(abup_uart_context[uart_num].hal), uart_num);
    uart_hal_set_sclk(&(abup_uart_context[uart_num].hal), uart_config->source_clk);
    uart_hal_set_baudrate(&(abup_uart_context[uart_num].hal), uart_config->baud_rate);
    uart_hal_set_parity(&(abup_uart_context[uart_num].hal), uart_config->parity);
    uart_hal_set_data_bit_num(&(abup_uart_context[uart_num].hal), uart_config->data_bits);
    uart_hal_set_stop_bits(&(abup_uart_context[uart_num].hal), uart_config->stop_bits);
    uart_hal_set_tx_idle_num(&(abup_uart_context[uart_num].hal), ABUP_UART_TX_IDLE_NUM_DEFAULT);
    uart_hal_set_hw_flow_ctrl(&(abup_uart_context[uart_num].hal), uart_config->flow_ctrl, uart_config->rx_flow_ctrl_thresh);

    uart_hal_rxfifo_rst(&(abup_uart_context[uart_num].hal));
    uart_hal_txfifo_rst(&(abup_uart_context[uart_num].hal));
    return ESP_OK;
}

esp_err_t abup_uart_intr_config(uart_port_t uart_num, const abup_uart_intr_config_t *intr_conf)
{
    UART_CHECK((uart_num < UART_NUM_MAX), "uart_num error", ESP_FAIL);
    UART_CHECK((intr_conf), "param null", ESP_FAIL);
    uart_hal_clr_intsts_mask(&(abup_uart_context[uart_num].hal), UART_LL_INTR_MASK);
    if(intr_conf->intr_enable_mask & UART_INTR_RXFIFO_TOUT) {
        uart_hal_set_rx_timeout(&(abup_uart_context[uart_num].hal), intr_conf->rx_timeout_thresh);
    } else {
        //Disable rx_tout intr
        uart_hal_set_rx_timeout(&(abup_uart_context[uart_num].hal), 0);
    }
    if(intr_conf->intr_enable_mask & UART_INTR_RXFIFO_FULL) {
        uart_hal_set_rxfifo_full_thr(&(abup_uart_context[uart_num].hal), intr_conf->rxfifo_full_thresh);
    }
    if(intr_conf->intr_enable_mask & UART_INTR_TXFIFO_EMPTY) {
        uart_hal_set_txfifo_empty_thr(&(abup_uart_context[uart_num].hal), intr_conf->txfifo_empty_intr_thresh);
    }
    uart_hal_ena_intr_mask(&(abup_uart_context[uart_num].hal), intr_conf->intr_enable_mask);

    return ESP_OK;
}

static uint32_t UART_ISR_ATTR uart_enable_tx_write_fifo(uart_port_t uart_num, const uint8_t *pbuf, uint32_t len)
{
    uint32_t sent_len = 0;

    if (UART_IS_MODE_SET(uart_num, UART_MODE_RS485_HALF_DUPLEX)) {
        uart_hal_set_rts(&(abup_uart_context[uart_num].hal), 0);
        uart_hal_ena_intr_mask(&(abup_uart_context[uart_num].hal), UART_INTR_TX_DONE);
    }
    uart_hal_write_txfifo(&(abup_uart_context[uart_num].hal), pbuf, len, &sent_len);

    return sent_len;
}


/**************************************************************/
int abup_uart_tx_chars(uart_port_t uart_num, const char* buffer, uint32_t len)
{
    UART_CHECK((uart_num < UART_NUM_MAX), "uart_num error", (-1));
    UART_CHECK(buffer, "buffer null", (-1));
    if(len == 0) {
        return 0;
    }
    int tx_len = 0;

    tx_len = (int)uart_enable_tx_write_fifo(uart_num, (const uint8_t *) buffer, len);

    return tx_len;
}


int abup_uart_read_bytes(uart_port_t uart_num)
{
    UART_CHECK((uart_num < UART_NUM_MAX), "uart_num error", (-1));
    size_t size = 0;
    
    while(1) {
        if (abup_timer_callback_cnt >= 15)
        {
            break;
        }
        else
        {
            uart_hal_read_rxfifo(&(abup_uart_context[uart_num].hal), abup_uart_obj[uart_num].rx_data_buf, &size);
        }
    }
    return size;
}

esp_err_t uart_get_buffered_data_len(uart_port_t uart_num, size_t* size)
{
    UART_CHECK((uart_num < UART_NUM_MAX), "uart_num error", ESP_FAIL);
    *size = abup_uart_obj[uart_num].rx_buffered_len;
    return ESP_OK;
}


esp_err_t abup_uart_driver_install(uart_port_t uart_num, int rx_buffer_size, int tx_buffer_size, int queue_size, int intr_alloc_flags)
{
    esp_err_t r = ESP_FAIL;
    UART_CHECK((uart_num < UART_NUM_MAX), "uart_num error", ESP_FAIL);
    //UART_CHECK((rx_buffer_size > SOC_UART_FIFO_LEN), "uart rx buffer length error", ESP_FAIL);
    //UART_CHECK((tx_buffer_size > SOC_UART_FIFO_LEN) || (tx_buffer_size == 0), "uart tx buffer length error", ESP_FAIL);
#if CONFIG_UART_ISR_IN_IRAM
    if ((intr_alloc_flags & ESP_INTR_FLAG_IRAM) == 0) {
        ESP_LOGI(ABUP_UART_TAG, "ESP_INTR_FLAG_IRAM flag not set while CONFIG_UART_ISR_IN_IRAM is enabled, flag updated");
        intr_alloc_flags |= ESP_INTR_FLAG_IRAM;
    }
#else
    if ((intr_alloc_flags & ESP_INTR_FLAG_IRAM) != 0) {
        ESP_LOGI(ABUP_UART_TAG, "ESP_INTR_FLAG_IRAM flag is set while CONFIG_UART_ISR_IN_IRAM is not enabled, flag updated");
        intr_alloc_flags &= ~ESP_INTR_FLAG_IRAM;
    }
#endif

    abup_uart_obj[uart_num].uart_num = uart_num;
    abup_uart_obj[uart_num].uart_mode = UART_MODE_UART;
    abup_uart_obj[uart_num].coll_det_flg = false;
    abup_uart_obj[uart_num].rx_always_timeout_flg = false;
	abup_uart_obj[uart_num].event_queue_size = queue_size;
    abup_uart_obj[uart_num].tx_ptr = NULL;
    abup_uart_obj[uart_num].tx_head = NULL;
    abup_uart_obj[uart_num].tx_len_tot = 0;
    abup_uart_obj[uart_num].tx_brk_flg = 0;
    abup_uart_obj[uart_num].tx_brk_len = 0;
    abup_uart_obj[uart_num].tx_waiting_brk = 0;
    abup_uart_obj[uart_num].rx_buffered_len = 0;
    abup_uart_obj[uart_num].rx_buffer_full_flg = false;
    abup_uart_obj[uart_num].tx_waiting_fifo = false;
    abup_uart_obj[uart_num].rx_ptr = NULL;
    abup_uart_obj[uart_num].rx_cur_remain = 0;
    abup_uart_obj[uart_num].rx_int_usr_mask = UART_INTR_RXFIFO_FULL | UART_INTR_RXFIFO_TOUT;
    abup_uart_obj[uart_num].rx_head_ptr = NULL;
    abup_uart_obj[uart_num].tx_buf_size = tx_buffer_size;
    //abup_uart_obj[uart_num].uart_select_notif_callback = NULL;

    abup_uart_intr_config_t abup_uart_intr = {
        .intr_enable_mask = ABUP_UART_INTR_CONFIG_FLAG,
        .rxfifo_full_thresh = ABUP_UART_FULL_THRESH_DEFAULT,
        .rx_timeout_thresh = ABUP_UART_TOUT_THRESH_DEFAULT,
        .txfifo_empty_intr_thresh = ABUP_UART_EMPTY_THRESH_DEFAULT,
    };
    abup_uart_module_enable(uart_num);
    uart_hal_disable_intr_mask(&(abup_uart_context[uart_num].hal), UART_LL_INTR_MASK);
    uart_hal_clr_intsts_mask(&(abup_uart_context[uart_num].hal), UART_LL_INTR_MASK);
    abup_uart_enable_rx_intr(uart_num);
    r=abup_uart_isr_register(uart_num, abup_uart_group_isr_callback, NULL);
    if (r!=ESP_OK) goto err;
    r=abup_uart_intr_config(uart_num, &abup_uart_intr);
    if (r!=ESP_OK) goto err;
    return r;

err:
    abup_uart_driver_delete(uart_num);
    return r;
}


//Make sure no other tasks are still using UART before you call this function
esp_err_t abup_uart_driver_delete(uart_port_t uart_num)
{
    UART_CHECK((uart_num < UART_NUM_MAX), "uart_num error", ESP_FAIL);

    uart_disable_rx_intr(uart_num);
    uart_disable_tx_intr(uart_num);

#if SOC_UART_SUPPORT_RTC_CLK

    uart_sclk_t sclk = 0;
    uart_hal_get_sclk(&(abup_uart_context[uart_num].hal), &sclk);
    if (sclk == UART_SCLK_RTC) {
        rtc_clk_disable(uart_num);
    }
#endif
    uart_module_disable(uart_num);
    return ESP_OK;
}


// Set UART mode
esp_err_t uart_set_mode(uart_port_t uart_num, uart_mode_t mode)
{
    UART_CHECK((uart_num < UART_NUM_MAX), "uart_num error", ESP_ERR_INVALID_ARG);
    if ((mode == UART_MODE_RS485_COLLISION_DETECT) || (mode == UART_MODE_RS485_APP_CTRL)
            || (mode == UART_MODE_RS485_HALF_DUPLEX)) {
        UART_CHECK((!uart_hal_is_hw_rts_en(&(abup_uart_context[uart_num].hal))),
                "disable hw flowctrl before using RS485 mode", ESP_ERR_INVALID_ARG);
    }

    uart_hal_set_mode(&(abup_uart_context[uart_num].hal), mode);
    if(mode ==  UART_MODE_RS485_COLLISION_DETECT) {
        // This mode allows read while transmitting that allows collision detection
        abup_uart_obj[uart_num].coll_det_flg = false;
        // Enable collision detection interrupts
        uart_hal_ena_intr_mask(&(abup_uart_context[uart_num].hal), UART_INTR_RXFIFO_TOUT
                                        | UART_INTR_RXFIFO_FULL
                                        | UART_INTR_RS485_CLASH
                                        | UART_INTR_RS485_FRM_ERR
                                        | UART_INTR_RS485_PARITY_ERR);
    }
    abup_uart_obj[uart_num].uart_mode = mode;

    return ESP_OK;
}

esp_err_t uart_set_rx_full_threshold(uart_port_t uart_num, int threshold)
{
    UART_CHECK((uart_num < UART_NUM_MAX), "uart_num error", ESP_ERR_INVALID_ARG);
    UART_CHECK((threshold < UART_RXFIFO_FULL_THRHD_V) && (threshold > 0),
        "rx fifo full threshold value error", ESP_ERR_INVALID_ARG);

    if (uart_hal_get_intr_ena_status(&(abup_uart_context[uart_num].hal)) & UART_INTR_RXFIFO_FULL) {
        uart_hal_set_rxfifo_full_thr(&(abup_uart_context[uart_num].hal), threshold);
    }

    return ESP_OK;
}

esp_err_t uart_set_tx_empty_threshold(uart_port_t uart_num, int threshold)
{
    UART_CHECK((uart_num < UART_NUM_MAX), "uart_num error", ESP_ERR_INVALID_ARG);
    UART_CHECK((threshold < UART_TXFIFO_EMPTY_THRHD_V) && (threshold > 0),
        "tx fifo empty threshold value error", ESP_ERR_INVALID_ARG);

    if (uart_hal_get_intr_ena_status(&(abup_uart_context[uart_num].hal)) & UART_INTR_TXFIFO_EMPTY) {
        uart_hal_set_txfifo_empty_thr(&(abup_uart_context[uart_num].hal), threshold);
    }

    return ESP_OK;
}

esp_err_t uart_set_rx_timeout(uart_port_t uart_num, const uint8_t tout_thresh)
{
    UART_CHECK((uart_num < UART_NUM_MAX), "uart_num error", ESP_ERR_INVALID_ARG);
    // get maximum timeout threshold
    uint16_t tout_max_thresh = uart_hal_get_max_rx_timeout_thrd(&(abup_uart_context[uart_num].hal));
    if (tout_thresh > tout_max_thresh) {
        ESP_LOGE(ABUP_UART_TAG, "tout_thresh = %d > maximum value = %d", tout_thresh, tout_max_thresh);
        return ESP_ERR_INVALID_ARG;
    }

    uart_hal_set_rx_timeout(&(abup_uart_context[uart_num].hal), tout_thresh);

    return ESP_OK;
}

esp_err_t uart_get_collision_flag(uart_port_t uart_num, bool* collision_flag)
{
    UART_CHECK((uart_num < UART_NUM_MAX), "uart_num error", ESP_ERR_INVALID_ARG);
    UART_CHECK((collision_flag != NULL), "wrong parameter pointer", ESP_ERR_INVALID_ARG);
    UART_CHECK((UART_IS_MODE_SET(uart_num, UART_MODE_RS485_HALF_DUPLEX)
                    || UART_IS_MODE_SET(uart_num, UART_MODE_RS485_COLLISION_DETECT)),
                    "wrong mode", ESP_ERR_INVALID_ARG);
    *collision_flag = abup_uart_obj[uart_num].coll_det_flg;
    return ESP_OK;
}

esp_err_t uart_set_wakeup_threshold(uart_port_t uart_num, int wakeup_threshold)
{
    UART_CHECK((uart_num < UART_NUM_MAX), "uart_num error", ESP_ERR_INVALID_ARG);
    UART_CHECK((wakeup_threshold <= UART_ACTIVE_THRESHOLD_V &&
                wakeup_threshold > ABUP_UART_MIN_WAKEUP_THRESH),
                "wakeup_threshold out of bounds", ESP_ERR_INVALID_ARG);

    uart_hal_set_wakeup_thrd(&(abup_uart_context[uart_num].hal), wakeup_threshold);

    return ESP_OK;
}

esp_err_t uart_get_wakeup_threshold(uart_port_t uart_num, int* out_wakeup_threshold)
{
    UART_CHECK((uart_num < UART_NUM_MAX), "uart_num error", ESP_ERR_INVALID_ARG);
    UART_CHECK((out_wakeup_threshold != NULL), "argument is NULL", ESP_ERR_INVALID_ARG);
    uart_hal_get_wakeup_thrd(&(abup_uart_context[uart_num].hal), (uint32_t *)out_wakeup_threshold);
    return ESP_OK;
}

esp_err_t uart_wait_tx_idle_polling(uart_port_t uart_num)
{
    UART_CHECK((uart_num < UART_NUM_MAX), "uart_num error", ESP_ERR_INVALID_ARG);
    while(!uart_hal_is_tx_idle(&(abup_uart_context[uart_num].hal)));
    return ESP_OK;
}

esp_err_t uart_set_loop_back(uart_port_t uart_num, bool loop_back_en)
{
    UART_CHECK((uart_num < UART_NUM_MAX), "uart_num error", ESP_ERR_INVALID_ARG);
    uart_hal_set_loop_back(&(abup_uart_context[uart_num].hal), loop_back_en);
    return ESP_OK;
}

void uart_set_always_rx_timeout(uart_port_t uart_num, bool always_rx_timeout)
{
    uint16_t rx_tout = uart_hal_get_rx_tout_thr(&(abup_uart_context[uart_num].hal));
    if (rx_tout) {
        abup_uart_obj[uart_num].rx_always_timeout_flg = always_rx_timeout;
    } else {
        abup_uart_obj[uart_num].rx_always_timeout_flg = false;
    }
}
